# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import functools
import logging
import sys
import uuid

import six

from dbnd._core.cli.click_utils import _help
from dbnd._core.configuration.config_readers import parse_and_build_config_store
from dbnd._core.configuration.config_value import ConfigValuePriority
from dbnd._core.configuration.environ_config import tracking_mode_context
from dbnd._core.configuration.pprint_config import pformat_config_store_as_table
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.log.config import configure_basic_logging
from dbnd._core.log.dbnd_log import set_verbose
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.tracking.schemas.tracking_info_run import ScheduledRunInfo
from dbnd._core.utils.click_tzdatetime import TZAwareDateTime
from dbnd._vendor import click
from dbnd_run.cli.service_auto_completer import completer
from dbnd_run.orchestration_bootstrap import dbnd_disable_databand_dags_loading
from dbnd_run.utils.user_code import load_user_modules


logger = logging.getLogger(__name__)


def _to_conf(kwargs):
    return {k: str(v) for k, v in kwargs.items() if v is not None}


class ConfigValueType(click.ParamType):
    def __init__(self):
        self.name = "config"

    def convert(self, value, param, ctx):
        from dbnd._core.utils import json_utils
        from targets.values.structure import _PARSABLE_PARAM_PREFIX

        value = value.strip()
        # support for --set '{ "a":2, "b":33}'
        if value[0] in _PARSABLE_PARAM_PREFIX:
            return json_utils.loads(value)

        name, sep, var = value.partition("=")
        return {name: var}


@click.command(name="run", context_settings=dict(help_option_names=[]))
@click.argument("task", autocompletion=completer.task(), required=False)
@click.option(
    "--module",
    "-m",
    multiple=True,
    help="Define a path of a module where DBND can search for a task or pipeline",
)
@click.option(
    "--set",
    "-s",
    "_sets",
    help="Set a configuration value (task_name.task_parameter=value)",
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.task_param(),
)
@click.option(
    "--set-config",
    "-c",
    "_sets_config",
    help="Set configuration a value (key=value). Example: --set run.name=my_run",
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.config_param(),
)
@click.option(
    "--extend",
    "-x",
    "_extend",
    help="extend configuration value. The configuration values must support extending. Example: --extend core.tracker='["
    '"debug"]\'',
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.config_param(),
)
@click.option(
    "--set-root",
    "-r",
    "_sets_root",
    help="Set or override a task parameter value (task_parameter=value)",
    type=ConfigValueType(),
    multiple=True,
    autocompletion=completer.root_param(),
)
@click.option(
    "--verbose", "-v", count=True, help="Make the logging output more verbose"
)
@click.option("--describe", is_flag=True, help="Describe the current run")
@click.option("--env", help="task environment: local/aws/aws_prod/gcp/prod")
@click.option("--parallel", is_flag=True, help="Run tasks in parallel")
@click.option("--conf-file", help="List of files to read from")
@click.option("--task-version", help="Task version directly affects the task signature")
@click.option("--project", help="Name of this Databand project")
@click.option("--name", help="Name of this run")
@click.option("--description", help="Description of this run")
@click.option(
    "--override",
    "-o",
    "_overrides",
    type=ConfigValueType(),
    multiple=True,
    help="Override configuration value (higher priority than any config source) key=value",
)
@click.option(
    "--help", "is_help", is_flag=True, help="Used for dynamic loading of modules"
)
@click.option(
    "--run-driver", "run_driver", type=uuid.UUID, help="Running in remote mode"
)
@click.option(
    "--override-run-uid",
    "override_run_uid",
    type=uuid.UUID,
    help="Use predefined uid for run. Run will be considered as a new run.",
)
@click.option("--task-name", "alternative_task_name", help="Name of this task")
@click.option("--job-name", "job_name", help="Job Name")
@click.option(
    "--scheduled-job-name",
    "-sjn",
    help="Associate this run with the scheduled job which will be created if needed",
)
@click.option(
    "--scheduled-date",
    "-sd",
    help="Should be used only when scheduled-job-name is set",
    type=TZAwareDateTime(),
)
@click.option("--interactive", is_flag=True, help="Run submission in blocking mode")
@click.option(
    "--submit-driver",
    "submit_driver",
    flag_value=1,
    help="Run driver at a remote engine",
)
@click.option(
    "--local-driver", "submit_driver", flag_value=0, help="Run the driver locally"
)
@click.option(
    "--submit-tasks",
    "submit_tasks",
    flag_value=1,
    help="Submit tasks from driver to a remote engine",
)
@click.option(
    "--no-submit-tasks",
    "submit_tasks",
    flag_value=0,
    help="Disable task submission, run tasks in the driver process.",
)
@click.option("--disable-web-tracker", help="Disable web tracking", is_flag=True)
@click.option("--interactive", is_flag=True, help="Run submission in blocking mode")
@click.option(
    "--open-web-tab", help="Open the tracker URL in the Databand UI", is_flag=True
)
@click.option(
    "--docker-build-tag",
    help="Define custom docker image tag for docker image that will be built",
)
@click.pass_context
def cmd_run(
    ctx,
    is_help,
    task,
    module,
    _sets,
    _sets_config,
    _sets_root,
    _overrides,
    _extend,
    verbose,
    describe,
    env,
    parallel,
    conf_file,
    task_version,
    project,
    name,
    description,
    run_driver,
    override_run_uid,
    alternative_task_name,
    job_name,
    scheduled_job_name,
    scheduled_date,
    interactive,
    submit_driver,
    submit_tasks,
    disable_web_tracker,
    open_web_tab,
    docker_build_tag,
):
    """
    Run a task or a DAG

    To see all available tasks use `dbnd show-tasks` (tab completion is available).
    `dbnd show-configs` will print all available configs.
    """
    if verbose:
        set_verbose()

    # double checking on bootstrap, as we can run from all kind of locations
    # usually we should be bootstraped already as we run from cli.
    dbnd_bootstrap(enable_dbnd_run=True)

    # make all these tasks discoverable from cli
    from dbnd import config
    from dbnd._core.context.databand_context import DatabandContext, new_dbnd_context
    from dbnd._core.utils.structures import combine_mappings
    from dbnd_run.tasks import basics  # noqa: F401

    dbnd_disable_databand_dags_loading(dbnd_config=config)

    task_registry = get_task_registry()

    # we need to do it before we are looking for the task cls
    load_user_modules(dbnd_config=config, modules=module)

    task_name = task
    # --verbose, --describe, --env, --parallel, --conf-file and --project
    # we filter out false flags since otherwise they will always override the config with their falseness
    main_switches = dict(
        databand=dict(
            verbose=verbose > 0,
            describe=describe,
            env=env,
            conf_file=conf_file,
            project=project,
        ),
        run_info=dict(name=name, description=description, is_archived=describe),
        run=dict(
            parallel=parallel,
            interactive=interactive,
            open_web_tracker_in_browser=open_web_tab,
            submit_driver=_nullable_flag(submit_driver),
            submit_tasks=_nullable_flag(submit_tasks),
        ),
        kubernetes=dict(docker_build_tag=docker_build_tag),
        task=dict(task_version=task_version),
        task_build=dict(verbose=True if verbose > 1 else None),
        core=dict(tracker_api="disabled" if disable_web_tracker else None),
    )

    main_switches = cleanup_empty_switches(main_switches)

    _sets = list(_sets)
    _sets_config = list(_sets_config)
    _sets_root = list(_sets_root)

    root_task_config = {}
    for _set in _sets_root:
        root_task_config = combine_mappings(left=root_task_config, right=_set)

    # remove all "first level" config values, assume that they are for the main task
    # add them to _sets_root
    for _set in _sets:
        for k, v in list(_set.items()):
            # so json-like values won't be included
            if "." not in k and isinstance(v, six.string_types):
                root_task_config[k] = v
                del _set[k]

    cmd_line_config = parse_and_build_config_store(
        source="cli", config_values=main_switches
    )
    # --set, --set-config
    if _sets:
        cmd_line_config.update(_parse_cli(_sets, source="--set"))
    if _sets_config:
        cmd_line_config.update(_parse_cli(_sets_config, source="--set-config"))
    if _extend:
        cmd_line_config.update(
            _parse_cli(_extend, source="--extend-config", extend=True)
        )
    if _overrides:
        cmd_line_config.update(
            _parse_cli(
                _overrides,
                source="--set-override",
                priority=ConfigValuePriority.OVERRIDE,
            )
        )

    # --set-root
    if root_task_config:
        task_cls = task_registry.get_task_cls(task_name)
        task_section = task_cls.task_definition.task_config_section
        # adding root task to configuration
        cmd_line_config.update(
            parse_and_build_config_store(
                config_values={task_section: root_task_config}, source="--set-root"
            )
        )

    # UPDATE CURRENT CONFIG with CLI values
    if cmd_line_config:
        if verbose:
            logger.info(
                "CLI config: \n%s", pformat_config_store_as_table(cmd_line_config)
            )
        config.set_values(cmd_line_config, source="cmdline")

    # initialize basic logging (until we get to the context logging
    if not config.getboolean("log", "disabled"):
        configure_basic_logging(None)

    scheduled_run_info = None
    if scheduled_job_name:
        scheduled_run_info = ScheduledRunInfo(
            scheduled_job_name=scheduled_job_name, scheduled_date=scheduled_date
        )

    # update completer
    if config.getboolean("databand", "completer"):
        tasks = task_registry.list_dbnd_task_classes()
        completer.refresh(tasks)

    # bootstrap and modules are loaded, we can load the task
    task_cls = None
    if task_name:
        task_cls = task_registry.get_task_cls(task_name)

    if not task_name:
        print_help(ctx, None)
        return

    if is_help:
        print_help(ctx, task_cls)
        return

    with tracking_mode_context(tracking=False), new_dbnd_context(
        name="run"
    ) as context:  # type: DatabandContext
        if context.settings.system.describe:
            # we want to print describe without triggering real run
            logger.info("Building main task '%s'", task_name)
            root_task = get_task_registry().build_dbnd_task(task_name)
            root_task.ctrl.describe_dag.describe_dag()
            # currently there is bug with the click version we have when using python 2
            # so we don't use the click.echo function
            # https://github.com/pallets/click/issues/564
            print("Task %s has been described!" % task_name)
            return root_task
        return context.dbnd_run_task(
            task_or_task_name=task_name,
            force_task_name=alternative_task_name,
            job_name=job_name or alternative_task_name or task_name,
            run_uid=run_driver or override_run_uid,
            existing_run=run_driver is not None,
            scheduled_run_info=scheduled_run_info,
            project=project,
        )


def cleanup_empty_switches(main_switches):
    # remove all None (so we use config values)
    main_switches = {k: _filter_none(**v) for k, v in main_switches.items()}
    # remove all empty sections
    main_switches = {k: v for k, v in main_switches.items() if v}
    return main_switches


def is_running_in_direct_db_mode(ctx):
    return "api" in ctx.settings.core.tracker and ctx.settings.core.tracker_api == "db"


def print_help(ctx, task_cls):
    from dbnd._core.parameter.parameter_definition import _ParameterKind

    formatter = ctx.make_formatter()
    cmd_run.format_help(ctx, formatter)
    if task_cls:
        dl = []
        for (param_name, param_obj) in task_cls.task_definition.task_param_defs.items():
            if param_obj.system or param_obj.kind == _ParameterKind.task_output:
                continue

            param_help = _help(param_obj.description)
            dl.append(("-r " + param_name, param_help))

        with formatter.section("Task"):
            formatter.write_dl(dl)
    click.echo(formatter.getvalue().rstrip("\n"), color=ctx.color)


def _parse_cli(configs, source, priority=None, extend=False):
    """
    Parse every item in configs , joining them into one big ConfigStore
    """
    config_values_list = [
        parse_and_build_config_store(
            config_values=c,
            source=source,
            priority=priority,
            auto_section_parse=True,
            extend=extend,
        )
        for c in configs
    ]
    return functools.reduce((lambda x, y: x.update(y)), config_values_list)


def _filter_none(**kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}


def _nullable_flag(flag):
    if flag is None:
        return None

    return bool(flag)


def dbnd_run_cmd_main(task, env=None, args=None):
    """A wrapper for dbnd_cmd_run with error handling."""
    dbnd_bootstrap(enable_dbnd_run=True)

    from dbnd import Task

    if isinstance(task, Task):
        task_str = task.get_full_task_family()
    else:
        task_str = task

    try:
        cmd_args = [task_str]
        if env is not None:
            cmd_args = cmd_args + ["--env", env]
        if args is not None:
            cmd_args = cmd_args + args

        from dbnd._core.cli.main import dbnd_cmd

        return dbnd_cmd("run", cmd_args)
    except KeyboardInterrupt:
        logger.error("Keyboard interrupt, exiting...")
        sys.exit(1)
    except Exception as ex:
        from dbnd._core.failures import get_databand_error_message

        msg, code = get_databand_error_message(ex=ex, args=sys.argv[1:])
        logger.error("dbnd cmd run failed with error: {}".format(msg))
        if code is not None:
            sys.exit(code)



# © Copyright Databand.ai, an IBM Company 2022

from typing import Type

from dbnd._core.configuration.dbnd_config import config
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import Config
from targets.target_config import TargetConfig, file, parse_target_config
from targets.values import get_value_type_of_type
from targets.values.version_value import VersionAlias, VersionStr


class OutputConfig(Config):
    """(Advanced) Databand's core task's output behaviour"""

    _conf__task_family = "output"

    path_task = parameter(description="Set the default path for every Task.")[str]
    path_prod_immutable_task = parameter(
        description="Set the format of the path to be used by Production Immutable tasks."
    )[str]

    hdf_format = (
        parameter.choices(["table", "fixed"])
        .help("Determine default format to save DataFrame to hdf.")
        .value("fixed")
    )

    deploy_id = parameter(
        default=VersionAlias.context_uid,
        description="Set deploy prefix to use for remote deployments.",
    )[VersionStr]

    # temporal location
    recheck_circle_dependencies = parameter(
        description="Recheck circle dependencies on every task creation, "
        "use it if you need to find a circle in your graph."
    ).value(False)

    def get_value_target_config(self, value_type):
        # type: (Type) -> TargetConfig

        type_handler = get_value_type_of_type(value_type)
        for possible_option in [str(type_handler), type_handler.config_name]:
            config_value = config.get_config_value(
                section=self._conf__task_family, key=possible_option
            )
            if config_value:
                return parse_target_config(config_value.value)
        return file.pickle



# © Copyright Databand.ai, an IBM Company 2022

import logging
import sys
import warnings

from dbnd._core.context.use_dbnd_run import (
    is_dbnd_orchestration_via_airflow_enabled,
    is_orchestration_mode,
)
from dbnd._core.errors import DatabandSystemError, friendly_error
from dbnd._core.errors.friendly_error import _band_call_str
from dbnd._core.parameter.parameter_value import ParameterFilters
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.task_build.task_signature import (
    build_signature,
    build_signature_from_values,
)
from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl
from dbnd._core.utils.basics.nested_context import nested
from dbnd._core.utils.basics.nothing import is_not_defined
from dbnd._core.utils.task_utils import to_targets
from dbnd._core.utils.traversing import traverse
from targets import Target
from targets.base_target import TargetSource


logger = logging.getLogger(__name__)


def _section(parameter):
    return "system" if parameter.system else "user"


class TaskRelations(TaskSubCtrl):
    def __init__(self, task):
        super(TaskRelations, self).__init__(task)
        from dbnd._core.task.task_mixin import _TaskCtrlMixin

        # map of required targets
        self.task_inputs = None
        self.task_outputs = None

        self.task_band_result = None  # real output of task_band

        assert isinstance(task, _TaskCtrlMixin)

    def initialize_relations(self):
        # STEP 0 - run band function
        self.initialize_band()

        # STEP 1 - calculate all inputs and _required
        try:
            self.task_inputs = self.initialize_required()
        except Exception:
            logger.warning(
                "Failed to calculate relationships for %s" % self.task_id, exc_info=True
            )
            self.task_inputs = {}
            if not self.task.task_is_dynamic:
                raise

        # STEP 2 ( now we have all inputs, we can calculate real signature)
        # support for two phase build
        # will be called from MetaClass

        params = self.params.get_params_serialized(ParameterFilters.SIGNIFICANT_INPUTS)

        if "user" in self.task_inputs:
            # TODO : why do we need to convert all "user side" inputs?
            # what if the input is insignificant?
            system_input = self.task_inputs.get("system")
            if system_input and "band" in system_input:
                band_input = system_input["band"]
                task_inputs_user_only = {
                    "user": self.task_inputs.get("user"),
                    "system": {"band": band_input},
                }
            else:
                task_inputs_user_only = {"user": self.task_inputs.get("user")}
            task_inputs_as_str = traverse(
                task_inputs_user_only,
                convert_f=str,
                filter_none=True,
                filter_empty=True,
            )

            if task_inputs_as_str is None:
                task_inputs_as_str = ""

            params.append(("_task_inputs", task_inputs_as_str))

        # IMPORTANT PART: we initialize task_id here again
        # after all values are calculated (all task_inputs are assigned)
        # we do it again, now we have all inputs calculated
        task = self.task
        task.task_signature_obj = build_signature(
            name=task.task_name,
            params=params,
            extra=task.task_definition.task_signature_extra,
        )
        task.task_id = "{}__{}".format(
            task.task_name, task.task_signature_obj.signature
        )

        # for airflow operator task handling:
        airflow_task_id_p = self.params.get_param("airflow_task_id")
        if airflow_task_id_p:
            self.task.task_id = self.task.airflow_task_id

        # STEP 3  - now let update outputs
        self.initialize_outputs()

        outputs_sig = self._get_outputs_to_sign()
        if outputs_sig:
            sig = build_signature_from_values("task_outputs", outputs_sig)
            task.task_outputs_signature_obj = sig
        else:
            task.task_outputs_signature_obj = task.task_signature_obj

    def _get_outputs_to_sign(self):
        outputs_to_sign = self.task_outputs_user
        from dbnd import PipelineTask

        if not isinstance(self.task, PipelineTask):
            outputs_to_sign = {
                k: v for k, v in outputs_to_sign.items() if k != "task_band"
            }
        return outputs_to_sign

    def initialize_band(self):
        # we run this code for orchestration only!
        if not is_orchestration_mode():
            return

        try:
            band_context = []
            if is_dbnd_orchestration_via_airflow_enabled():
                from dbnd_run.airflow.dbnd_task_executor.airflow_operators_catcher import (
                    get_databand_op_catcher_dag,
                )

                band_context.append(get_databand_op_catcher_dag())

            original_param_values = []
            for param_value in self.task.task_params.get_param_values(
                ParameterFilters.OUTPUTS
            ):
                if param_value.name == "task_band" or isinstance(
                    param_value.parameter, FuncResultParameter
                ):
                    continue
                original_param_values.append((param_value, param_value.value))

            with nested(*band_context):
                band = self.task.band()
                # this one would be normalized
                self.task._task_band_result = band
            self.task_band_result = band  # real value

            from dbnd import PipelineTask

            if isinstance(self.task, PipelineTask):
                # after .band has finished, all user outputs of the .band should be defined
                for param_value, _ in original_param_values:
                    # we want to validate only user facing parameters
                    # they should have assigned values by this moment,
                    # pipeline task can not have None outputs, after band call
                    if param_value.parameter.system:
                        continue
                    if is_not_defined(param_value.value):
                        raise friendly_error.task_build.pipeline_task_has_unassigned_outputs(
                            task=self.task, param=param_value.parameter
                        )

            # now let's normalize if user has changed outputs
            for param_value, original_value in original_param_values:
                if param_value.value is original_value:
                    continue

                try:
                    from dbnd._core.utils.task_utils import to_targets

                    normalized_value = to_targets(param_value.value)
                    param_value.update_param_value(normalized_value)
                except Exception as ex:
                    raise friendly_error.task_build.failed_to_assign_param_value_at_band(
                        ex, param_value.parameter, param_value.value, self.task
                    )

        except Exception as ex:
            logger.warning(
                self.task.ctrl.banner(
                    msg="Failed to run %s" % _band_call_str(self.task),
                    color="red",
                    exc_info=sys.exc_info(),
                )
            )

            if self.task.task_decorator:
                # just re-raise, we already have an error from the "run" function
                raise

            raise friendly_error.task_build.failed_to_call_band(ex, self.task)

    def _get_all_child_pipelines(self):
        from dbnd import PipelineTask

        child_pipelines = []
        for child in self.task.descendants.children:
            child_task = self.get_task_by_task_id(child)
            if isinstance(child_task, PipelineTask):
                child_pipelines.append(child_task)
        return child_pipelines

    def initialize_required(self):
        # regular requirements -- just all inputs
        inputs = {"user": {}, "system": {}}

        # we take all parameters that are inputs (not outputs)
        # however Primitive parameters are inputs only if they are Target (deferred)
        #           if isinstance(p, _TargetParameter) or isinstance(value, Target)

        for p, value in self.params.get_params_with_value(ParameterFilters.INPUTS):
            if value is None:
                continue
            value = traverse(
                value, convert_f=_find_target, filter_none=True, filter_empty=True
            )

            if not value:
                continue

            inputs[_section(p)][p.name] = value

        def _extend_system_section(key, extra):
            if not extra:
                return
            inputs["system"][key] = extra

        from dbnd import PipelineTask

        if isinstance(self.task, PipelineTask):
            task_output_values = {}
            for p, value in self.params.get_params_with_value(
                ParameterFilters.USER_OUTPUTS
            ):
                if p.name == "task_band" or isinstance(p, FuncResultParameter):
                    continue

                # band outputs are going to be required as inputs!
                # @pipeline can run only when all of it's "outputs" are ready
                task_output_values[p.name] = value

            _extend_system_section("band", task_output_values)

        # find all child pipelines and make them upstreams to the task
        _extend_system_section(
            "pipelines", {p.task_id: p for p in self._get_all_child_pipelines()}
        )
        # now may be user still use function _requires - so let add that to dependencies
        _extend_system_section("required", self.task._requires())

        return to_targets(inputs)

    def initialize_outputs(self):
        """
        The default output that this Task produces. Use outputs! Override only if you are writing "base" class
        """
        task = self.task

        outputs = {"user": {}, "system": {}}

        for p, value in self.params.get_params_with_value(ParameterFilters.OUTPUTS):
            if is_not_defined(value):
                value = p.build_output(task=task)
                setattr(self.task, p.name, value)

            if isinstance(p, FuncResultParameter):
                continue

            value = traverse_and_set_target(value, p._target_source(self.task))
            outputs[_section(p)][p.name] = value

        custom_outputs = self.task._output()
        if custom_outputs:
            if outputs["user"]:
                warnings.warn(
                    "Task %s has custom outputs in _output() function, all other outputs will be removed: %s"
                    % (task, outputs["user"]),
                    stacklevel=2,
                )
                outputs["user"] = custom_outputs

        # take ownership of all outputs and clean it, just in case
        # usually all outputs are assigned to task

        # just in case we have some "outputs" with Tasks
        outputs = to_targets(outputs)
        self.task_outputs = traverse_and_set_target(
            outputs, target_source=TargetSource(task_id=self.task_id)
        )
        # if not output.get('user', None):
        #
        #     raise DatabandBuildError("Task %s doesn't have any outputs defined." % task,
        #                              help_msg="1. Define output fields in Task class definition. \n"
        #                                       "2. If you have overrided Task._output() function "
        #                                       "it should have non empty return value")

    @property
    def task_inputs_user(self):
        return self.task_inputs.get("user", {})

    @property
    def task_inputs_system(self):
        return self.task_inputs.get("system", {})

    @property
    def task_outputs_user(self):
        return self.task_outputs.get("user", {})

    @property
    def task_outputs_system(self):
        return self.task_outputs.get("system", {})


def _find_target(target):
    if target is None:
        return target

    if isinstance(target, Target):
        return target
    return None


def traverse_and_set_target(target, target_source):
    return traverse(
        target, convert_f=lambda t: __set_target(target=t, target_source=target_source)
    )


def __set_target(target, target_source):
    if not target:
        return target

    if not isinstance(target, Target):
        raise DatabandSystemError(
            "Expected target object, got '%s' : %s" % (type(target), target)
        )
    if not target.source:
        target.source = target_source
    return target


def as_task(task_or_result):
    """Extracts task from `task_or_result`."""
    from dbnd import Task

    if isinstance(task_or_result, Target):
        return task_or_result.source_task
    if isinstance(task_or_result, Task):
        return task_or_result
    raise DatabandSystemError("Can not extract task from %s" % task_or_result)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors import DatabandConfigError


def local_engine_not_accept_remote_jobs(env, to_cluster):
    return DatabandConfigError(
        "cluster  '{to_cluster}' can't accept remote jobs".format(
            to_cluster=to_cluster
        ),
        help_msg="Check your execution configuration at {}".format(env),
    )


def kubernetes_with_non_compatible_engine(task_engine):
    return DatabandConfigError(
        "Config env.task_executor_type expected to be airflow_kubernetes for {}".format(
            task_engine.task_name
        )
    )


class DatabandKubernetesError(DatabandConfigError):
    pass


class KubernetesImageNotFoundError(DatabandKubernetesError):
    pass


class KubernetesRunningPodConditionFailure(DatabandKubernetesError):
    pass


class KubernetesPodConfigFailure(DatabandKubernetesError):
    pass


def kubernetes_image_not_found(image_name, message, long_msg):
    return KubernetesImageNotFoundError(
        "Failed to start Kubernetes pod because the configured image (%s) could not be pulled by Kubernetes: %s"
        % (image_name, message),
        help_msg="Make sure you have built and pushed your image. "
        "If the image is in a private repository make sure you "
        "configured image pull secrets for it in the Kubernetes cluster "
        "and configured image_pull_secrets in the Kubernetes engine config. Details %s"
        % long_msg,
    )


def kubernetes_running_pod_fails_on_condition(condition, pod_name):
    return KubernetesRunningPodConditionFailure(
        "Pod %s is failing at Running phase: %s" % (pod_name, condition),
        help_msg="The pod is failing. Main reason for that is preemptible nodes, "
        "you can disable this check at `kubernetes.check_running_pod_errors=False`. "
        "See pod log for more details",
    )


def kubernetes_pod_config_error(kub_message):
    help_msg = ""
    if "databand-secrets" in kub_message:
        help_msg = (
            "by default Databand assumes a secret exists in Kubernetes with the name databand-secrets "
            "containing the db connection string and airflow fernet key. "
            "Either create the secret (see the kubernetes deployment script) or override the secrets "
            "property in your Kubernetes engine config section"
        )
    elif kub_message.startswith("secret"):
        help_msg = "Check that the secret exsits in the cluster or remove it from the engine config"

    return KubernetesPodConfigFailure(
        "Failed to start Kubernetes pod because of a configuration error: %s"
        % kub_message,
        help_msg=help_msg,
    )


def kubernetes_pod_unschedulable(kub_message, extra_help=None):
    help_msg = ""
    if "taints" in kub_message:
        help_msg = "Either remove taints from at least one of the Kubernetes nodes or add tolerations to the Kubernetes engine config."

    if extra_help:
        help_msg = help_msg + "\n" + extra_help

    return DatabandConfigError(
        "Failed to start Kubernetes pod because it couldn't be scheduled. Reason: %s"
        % kub_message,
        help_msg=help_msg,
    )


def no_tag_on_no_build():
    return DatabandConfigError(
        "You are running with 'docker_build=False', however, 'container_tag' is not defined",
        help_msg="Please configure 'container_ta'g or change 'docker_build' to 'True'.",
    )


def failed_to_connect_to_cluster(in_cluster_value, exc):
    return DatabandConfigError(
        "Could not connect to kubernetes cluster! Exception: %s" % (exc,),
        help_msg="in-cluster is set to '%s'. Are you running %s cluster?"
        % (in_cluster_value, "inside" if in_cluster_value else "outside"),
    )


def failed_to_load_config_file(exc):
    return DatabandConfigError(
        "Could not connect to kubernetes cluster! Exception: %s" % (exc,),
        help_msg="there is probably an error in your configurations file.",
    )



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.constants import ApacheBeamClusterType, CloudType, EnvLabel
from dbnd._core.errors import friendly_error
from dbnd._core.parameter.constants import ParameterScope
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task.config import Config
from targets import DirTarget


logger = logging.getLogger(__name__)

task_env_param = parameter(scope=ParameterScope.children)


class EnvConfig(Config):
    """Databand's environment configuration"""

    _conf__task_family = "env"
    cloud_type = parameter(
        description="Set which cloud type is to be used. E.g. gcp, aws, etc."
    )[str]

    env_label = parameter(
        default=EnvLabel.dev,
        description="Set the environment type to be used. E.g. dev, int, prod.",
    )[
        str
    ]  # label

    production = parameter(
        description="This indicates that the environment is production."
    ).value(False)

    conn_id = parameter(default=None, description="Set the cloud connection settings.")[
        str
    ]

    # MAIN OUTPUT FOLDER
    root = parameter(description="Determine the main data output location.").folder[
        DirTarget
    ]

    # DATABAND SYSTEM FOLDERS
    dbnd_root = parameter(
        description="Determine DBND system output location."
    ).output.folder(default=None)[DirTarget]
    dbnd_local_root = parameter(
        description="Set DBND home for the local engine environment."
    ).output.folder()[DirTarget]
    dbnd_data_sync_root = parameter(
        description="Set rooted directory for target syncing against a remote engine."
    ).output.folder()[DirTarget]

    # execution
    local_engine = parameter(
        default="local_machine_engine",
        description="Set which engine will be used for local execution",
    )[str]
    remote_engine = parameter(
        description="Set the remote engine for the execution of driver/tasks"
    ).none[str]

    submit_driver = parameter(
        description="Enable submitting driver to `remote_engine`."
    ).none[bool]
    submit_tasks = parameter(
        description="Enable submitting tasks to remote engine one by one."
    ).none[bool]

    # properties that will affect "task-env" section
    spark_config = task_env_param.help(
        "Determine the Spark Configuration settings"
    ).value("spark")

    spark_engine = task_env_param.help(
        "Set the cluster engine to be used. E.g. local, emr (aws), dataproc (gcp), etc."
    ).value("spark_local")

    hdfs = task_env_param.help("Set the Hdfs cluster configuration settings").value(
        "hdfs_knox"
    )

    beam_config = task_env_param.help(
        "Set the Apache Beam configuration settings"
    ).value("beam")
    beam_engine = task_env_param.help(
        "Set the Apache Beam cluster engine. E.g. local or dataflow."
    ).value(ApacheBeamClusterType.local)

    docker_engine = task_env_param.help(
        "Set the Docker job engine, e.g. docker or aws_batch"
    ).value("docker")

    def _initialize(self):
        super(EnvConfig, self)._initialize()
        try:
            self.dbnd_root = self.dbnd_root or self.root.folder("dbnd")

            if not self.dbnd_local_root:
                if not self.dbnd_root.is_local():
                    raise friendly_error.config.dbnd_root_local_not_defined(self.name)
                self.dbnd_local_root = self.dbnd_root
        except Exception as e:
            raise friendly_error.task_build.failed_to_access_dbnd_home(
                self.dbnd_root, e
            )
        self.dbnd_local_root__build = self.dbnd_local_root.folder("build")

        if not self.dbnd_data_sync_root:
            self.dbnd_data_sync_root = self.dbnd_root.folder("sync")

        if self.submit_driver is None:
            self.submit_driver = bool(self.remote_engine)

        if self.submit_tasks is None:
            self.submit_tasks = bool(self.remote_engine)

    @property
    def name(self):
        return self.task_name

    @property
    def cloud_type(self):
        return self.task_family

    def prepare_env(self):
        pass


class LocalEnvConfig(EnvConfig):
    """
    Local environment configuration section
    """

    _conf__task_family = CloudType.local



from dbnd._core.cli.cmd_show import _list_tasks
from dbnd._core.configuration.dbnd_config import config as dbnd_config

# © Copyright Databand.ai, an IBM Company 2022
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._vendor import click
from dbnd_run.cli.service_auto_completer import completer
from dbnd_run.utils.user_code import load_user_modules


@click.command()
@click.argument("search", default="", autocompletion=completer.task())
@click.option("--module", "-m", help="Used for dynamic loading of modules")
@click.pass_context
def show_tasks(ctx, module, search):
    """Show and search tasks"""

    dbnd_bootstrap(enable_dbnd_run=True)

    load_user_modules(dbnd_config, modules=module)
    _list_tasks(ctx, search, is_config=False)



# © Copyright Databand.ai, an IBM Company 2022

import logging
import uuid

from dbnd._core.utils.basics.memoized import cached
from dbnd._core.utils.project.project_fs import abs_join, relative_path


logger = logging.getLogger(__name__)
_airflow_lib_home_default = relative_path(__file__)


def dbnd_airflow_path(*path):
    return abs_join(_airflow_lib_home_default, *path)


def create_airflow_pool(pool_name):
    from airflow.models import Pool
    from airflow.utils.db import create_session

    print("Creating Airflow pool '%s'" % pool_name)
    with create_session() as session:
        if session.query(Pool.pool).filter(Pool.pool == pool_name).scalar() is not None:
            return

        # -1 so we have endless pool
        dbnd_pool = Pool(pool=pool_name, slots=-1)
        session.merge(dbnd_pool)
        session.commit()


@cached()
def get_airflow_instance_uid():
    """used to distinguish between jobs of different airflow instances"""
    import airflow

    db_url = airflow.settings.Session.bind.engine.url
    db_str = "{}:{}/{}".format(db_url.host, db_url.port, db_url.database)
    airflow_instance_uid = uuid.uuid5(uuid.NAMESPACE_URL, db_str)
    return str(airflow_instance_uid)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import Task, output, parameter, project_path
from dbnd._core.current import try_get_databand_context
from dbnd_run.tasks.py_distribution.build_distribution import (
    _create_temp_working_dir,
    generate_project_whl,
    generate_third_party_deps,
    zip_dir,
)
from targets import FileTarget
from targets.types import PathStr


logger = logging.getLogger(__name__)


class ProjectWheelFile(Task):
    task_is_system = True
    package_dir = parameter(
        default=project_path(),
        description="Full path to the directory of the package that contains setup.py",
    )[PathStr]

    requirements_file = parameter(
        description="Full path to the requirements file (if any)", default=None
    )[PathStr]

    wheel_file = output.with_format("zip")

    def run(self):
        local_wheel_file_zip = (
            self.current_task_run_executor.local_task_run_root.partition(
                name="project_wheel_file.zip"
            )
        )  # type: FileTarget
        local_wheel_file_zip.mkdir_parent()
        # we need a temp folder to create our wheel file,
        # pip build doesn't take "output" parameter
        with _create_temp_working_dir() as tmp_build_dir:
            logger.info(
                "Started building bdist_zip file %s at %s, package_dir=%s",
                self.wheel_file,
                tmp_build_dir,
                self.package_dir,
            )

            # generate main file, we can find a package version and name from it if required
            generate_project_whl(self.package_dir, tmp_build_dir)
            if self.requirements_file is not None:
                generate_third_party_deps(self.requirements_file, tmp_build_dir)

            # zip everything into one file before we copy
            zip_dir(local_wheel_file_zip, tmp_build_dir)

            self.wheel_file.copy_from_local(local_wheel_file_zip)
            logger.info(
                "Successfully published project wheel zip file at %s", self.wheel_file
            )

    @classmethod
    def build_project_wheel_file_task(cls):
        # Use current_context_uid to make sure this task is going to be run only once per pipeline
        # Constant task_target_date so the signature won't change if user changes task_target_date parameter.
        fat_wheel_task = cls(
            # we need it to run every time we "rerun" the pipeline
            task_version=try_get_databand_context().current_context_uid,
            # we don't want to inherit from parent task, as they might have different target_dates
            task_target_date="today",
        )
        return fat_wheel_task



# -*- coding: utf-8 -*-
# © Copyright Databand.ai, an IBM Company 2022

"""This package is deprecated! Please use the dbnd package instead!"""

from dbnd import (
    Config,
    DataSourceTask,
    PipelineTask,
    PythonTask,
    Task,
    as_task,
    auto_namespace,
    band,
    config,
    config_deco,
    current,
    data,
    dbnd_context,
    dbnd_handle_errors,
    dbnd_tracking_start,
    dbnd_tracking_stop,
    hookimpl,
    namespace,
    new_dbnd_context,
    output,
    override,
    parameter,
    pipeline,
    task,
)
from dbnd_run.tasks import basics
from targets import _set_patches


__all__ = [
    "override",
    "config",
    "config_deco",
    "new_dbnd_context",
    "dbnd_context",
    "band",
    "pipeline",
    "task",
    "dbnd_handle_errors",
    "hookimpl",
    "dbnd_tracking_start",
    "dbnd_tracking_stop",
    "Config",
    "DataSourceTask",
    "PipelineTask",
    "PythonTask",
    "Task",
    "current",
    "auto_namespace",
    "namespace",
    "as_task",
    "data",
    "output",
    "parameter",
    "basics",
    "_set_patches",
]



# © Copyright Databand.ai, an IBM Company 2022

import logging
import time
import traceback


logger = logging.getLogger(__name__)


def remove_listener_by_name(target, identifier, name):
    """
    removes already registered sqlalchemy listener
    use this one only if fn pointer is not accessable (inner function)
    otherwise use regular remove from event api

    Example:
        from airflow import settings
        target = settings.engine
        remove_listener_by_name(target, "engine_connect", "ping_connection")
    """
    import ctypes

    from sqlalchemy import event

    all_keys = list(event.registry._key_to_collection.items())
    for key, values in all_keys:

        if key[0] != id(target):
            continue

        if identifier != key[1]:
            continue

        fn = ctypes.cast(key[2], ctypes.py_object).value  # get function by id
        if fn.__name__ != name:
            continue

        event.remove(target, identifier, fn)


def get_calling_line():
    code = "unknown"
    for (file_path, val1, val2, line_contents) in traceback.extract_stack():
        if "airflow" not in file_path:
            continue
        if (
            "utils/sqlalchemy.py" in file_path
            or "utils/db.py" in file_path
            or "db_utils" in file_path
        ):
            continue
        code = str((file_path, val1, val2))
    return code


def profile_before_cursor_execute(conn, cursor, statement, *_):
    conn.info.setdefault("query_start_time", []).append(time.time())
    logger.debug("Start Query: %s", statement)


def profile_after_cursor_execute(conn, cursor, statement, parameters, *_):
    total = time.time() - conn.info["query_start_time"].pop(-1)
    logger.debug(
        "Query Complete! %s  \n--> %f seconds\nPARAMS: %s", statement, total, parameters
    )


def airflow_tables_to_dump():
    from airflow import jobs as af_jobs, models as af_models

    return (
        # dbnd_dag -> dag_id
        # dbnd_airflow_models.DbndAirflowDagModel,
        # dbnd_dag_run -> user, [cmd_line]
        # dag_run -> dag_id,
        af_models.DagRun,
        af_models.TaskInstance,
        # dbnd_task_run -> created_by_task_id (__XXX), created_by_dag_id, task_name
        # dbnd_task_run_metrics -> name,
        # dag -> dag_id, [fileloc]
        af_models.DagModel,
        # job -> dag_id, hostname, unixname
        af_jobs.BaseJob,
    )


def airflow_sql_conn_url():
    try:
        from sqlalchemy.engine.url import make_url
    except:
        return "`pip install sqlalchemy` in order to get sql db url"

    return repr(make_url(airlow_sql_alchemy_conn()))


def airlow_sql_alchemy_conn():
    from airflow.configuration import conf as airflow_conf

    return airflow_conf.get("core", "sql_alchemy_conn")



# © Copyright Databand.ai, an IBM Company 2022

import json
import logging
import os

from functools import wraps

from dbnd._vendor.pygtrie import CharTrie


class FastSingletonContext(object):
    _instance = None

    @classmethod
    def try_instance(cls, *args, **kwargs):
        """Singleton get or create"""
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    @classmethod
    def renew_instance(cls, *args, **kwargs):
        cls._instance = None
        return cls.try_instance(*args, **kwargs)


class PrefixStore(object):
    def __init__(self, path):
        self._path = path
        self._indices = dict()
        self._objs = dict()

    def set(self, obj_id, obj):
        self._objs[obj_id] = obj

    def index(self, index_name, obj_ref, obj_id):
        if index_name not in self._indices:
            self._indices[index_name] = CharTrie()

        self._indices[index_name][obj_ref] = obj_id

    def search(self, indices_names, prefix):
        for index_name in indices_names:
            if index_name not in self._indices:
                continue

            index = self._indices[index_name]
            if prefix not in index and not index.has_subtrie(prefix):
                continue

            entries = index.items(prefix)
            if len(entries) == 0:
                continue

            return {obj_ref: self._objs[obj_id] for obj_ref, obj_id in entries}
        return dict()

    def load(self):
        cache = json.load(open(self._path, "r"))

        # see https://github.com/google/pygtrie/issues/9
        indices = dict()
        for index_name, index_val in cache["indices"].items():
            indices[index_name] = CharTrie()
            indices[index_name]._root.__setstate__(index_val)

        self._indices = indices
        self._objs = cache["objs"]

    def save(self):
        cache = dict()

        cache["indices"] = dict()

        # see https://github.com/google/pygtrie/issues/9
        for index_name, index_val in self._indices.items():
            cache["indices"][index_name] = index_val._root.__getstate__()

        cache["objs"] = self._objs

        try:
            cache_dir = os.path.dirname(self._path)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir)
            json.dump(cache, open(self._path, "w"))
        except Exception as ex:
            logging.error(
                "Failed to save autocompletion store at %s: %s", self._path, ex
            )


def no_errors(ret=None):
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception:
                return ret

        return wrapped

    return decorator



# © Copyright Databand.ai, an IBM Company 2022
import contextlib
import datetime
import logging
import random
import typing
import warnings

from contextlib import contextmanager
from typing import Any, Dict, Optional

from dbnd._core.configuration.dbnd_config import config as dbnd_config
from dbnd._core.constants import (
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    OutputMode,
    ParamValidation,
    TaskEssence,
    _TaskDbndRun,
    _TaskParamContainer,
)
from dbnd._core.current import get_databand_run, try_get_current_task_run
from dbnd._core.errors.friendly_error.task_build import incomplete_output_found_for_task
from dbnd._core.failures import dbnd_handle_errors
from dbnd._core.log.logging_utils import TaskContextFilter
from dbnd._core.parameter.constants import ParameterScope
from dbnd._core.parameter.parameter_builder import output, parameter
from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import ParameterFilters
from dbnd._core.task.task_mixin import _TaskCtrlMixin
from dbnd._core.task.task_with_params import _TaskWithParams
from dbnd._core.task_build.task_context import task_context
from dbnd._core.task_ctrl.task_ctrl import _BaseTaskCtrl
from dbnd._core.utils.basics.nested_context import nested
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._core.utils.traversing import flatten, traverse_to_str
from dbnd_run import errors
from dbnd_run.run_executor_engine.local_task_executor import topological_sort
from dbnd_run.run_settings import RunSettings
from dbnd_run.run_settings.env import EnvConfig
from dbnd_run.task_ctrl.task_output_builder import calculate_path
from targets import InMemoryTarget, target
from targets.target_config import TargetConfig, folder
from targets.values import get_value_type_of_obj
from targets.values.version_value import VersionStr


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.task_build.task_cls__call_state import TaskCallState
    from dbnd._core.task_ctrl.task_dag import _TaskDagNode
    from dbnd._core.task_run.task_run import TaskRun
    from dbnd_run.task_ctrl.task_run_executor import TaskRunExecutor

DEFAULT_CLASS_VERSION = ""

logger = logging.getLogger(__name__)

system_passthrough_param = parameter.system(scope=ParameterScope.children)


class Task(_TaskWithParams, _TaskCtrlMixin, _TaskParamContainer, _TaskDbndRun):
    """
    This is the base class of all dbnd Tasks, the base unit of work in databand.

    A dbnd Task describes a unit or work.

    A ``run`` method must be present in a subclass

    Each ``parameter`` of the Task should be declared as members::

        class MyTask(dbnd.Task):
            count = dbnd.parameter[int]
            second_param = dbnd.parameter[str]
    """

    _conf_confirm_on_kill_msg = None  # get user confirmation on task kill if not empty
    _conf__require_run_dump_file = False

    _task_band_result = output(default=None, system=True)
    _meta_output = output(
        system=True,
        output_name="meta",
        output_ext="",
        target_config=folder,
        significant=False,
        description="Location of all internal outputs (e.g. metrics)",
    )
    task_band = output.json(output_name="band", system=True)

    task_enabled = system_passthrough_param(default=True)[bool]
    task_enabled_in_prod = system_passthrough_param(default=True)[bool]
    validate_no_extra_params = ParamValidation.error

    # for permanent bump of task version use Task.task_class_version
    task_version = parameter(
        default="1",
        description="task version, directly affects task signature ",
        scope=ParameterScope.children,
    )[VersionStr]

    task_class_version = parameter.value(
        default=DEFAULT_CLASS_VERSION,
        system=True,
        description="task code version, "
        "use while you want persistent change in your task version",
    )

    task_env = parameter.value(
        default="local",
        description="task environment name",
        scope=ParameterScope.children,
    )[EnvConfig]

    task_target_date = parameter(
        default="today",
        description="task data target date",
        scope=ParameterScope.children,
    )[datetime.date]

    task_airflow_op_kwargs = parameter.system(
        default=None, description="airflow operator kwargs"
    )[Dict[str, object]]

    task_config = parameter.system(empty_default=True)[Dict]
    task_is_system = parameter.system(default=False)[bool]

    task_in_memory_outputs = system_passthrough_param(
        default=False, description="Store all task outputs in memory"
    )[bool]

    task_output_path_format = system_passthrough_param(
        default=None, description="Format string used to generate task output paths"
    )[str]

    task_is_dynamic = system_passthrough_param(
        default=False,
        scope=ParameterScope.children,
        description="task was executed from within another task",
    )[bool]

    # for example: if task.run doesn't have access to databand, we can't run runtime tasks
    task_supports_dynamic_tasks = parameter.system(
        default=True, description="indicates if task can run dynamic databand tasks"
    )[bool]

    task_retries = parameter.system(
        default=0,
        description="Total number of attempts to run the task. So task_retries=3 -> task can fail 3 times before we give up",
    )[int]

    task_retry_delay = parameter.system(
        default="15s",
        description="timedelta to wait before retrying a task. Example: 5s",
    )[datetime.timedelta]

    task_essence = TaskEssence.ORCHESTRATION

    def __init__(self, **kwargs):
        super(Task, self).__init__(**kwargs)

        # used to communicate return value of "user function"
        self._dbnd_call_state = None  # type: Optional[TaskCallState]
        self.ctrl = TaskCtrl(self)

    def band(self):
        """
        Please, do not override this function only in Pipeline/External tasks!

        We do all wiring work in Meta classes only.
        Our implementation should never be coupled to code!
        """
        return

    def run(self):
        """
        The task run method, to be overridden in a subclass.

        See :ref:`Task.run`
        """
        pass  # default impl

    @property
    def task_outputs(self):
        """
        The output that this Task produces.

        The output of the Task determines if the Task needs to be run--the task
        is considered finished iff the outputs all exist.
        """
        return self.ctrl.relations.task_outputs_user

    @property
    def task_dag(self):
        # type: (...)->_TaskDagNode
        return self.ctrl.task_dag

    @property
    def descendants(self):
        return self.ctrl.descendants

    def _complete(self):
        """
        If the task has any outputs, return ``True`` if all outputs exist. Otherwise, return ``False``.

        However, you may freely override this method with custom logic.
        """
        # we check only user side task outputs
        # all system tasks outputs are not important (if the exists or not)
        # user don't see them
        outputs = [
            o for o in flatten(self.task_outputs) if not o.config.overwrite_target
        ]
        if len(outputs) == 0:
            if not self.task_band:
                warnings.warn(
                    "Task %r without outputs has no custom complete() and no task band!"
                    % self,
                    stacklevel=2,
                )
                return False
            else:
                return self.task_band.exists()

        incomplete_outputs = [str(o) for o in outputs if not o.exists()]

        num_of_incomplete_outputs = len(incomplete_outputs)

        if 0 < num_of_incomplete_outputs < len(outputs):
            complete_outputs = [str(o) for o in outputs if o.exists()]
            exc = incomplete_output_found_for_task(
                self.task_name, complete_outputs, incomplete_outputs
            )

            if self.run_settings.run.validate_task_outputs_on_build:
                raise exc
            else:
                logger.warning(str(exc))

        return num_of_incomplete_outputs == 0

    @property
    def current_task_run(self):
        # type: ()->TaskRun
        return get_databand_run().get_task_run(self.task_id)

    @property
    def current_task_run_executor(self):
        # type: ()->TaskRunExecutor
        return self.current_task_run.task_run_executor

    @property
    def run_settings(self) -> RunSettings:
        return self.dbnd_context.run_settings

    def _output(self):
        """
        The default output that this Task produces.

        Use outputs! Override only if you are writing "base" class.
        """
        return NOTHING

    def _requires(self):
        """
        Override in "template" tasks which themselves are supposed to be subclassed.

        Must return an iterable which, among others, contains the _requires() of
        the superclass.
        """

    def _task_submit(self):
        """Task submission logic, by default we just call -> ``_task_run()`` -> ``run()``."""
        return self._task_run()

    def _task_run(self):
        # bring all relevant files
        task_run_executor = self.current_task_run.task_run_executor
        task_run_executor.sync_local.sync_pre_execute()
        param_values = self.task_params.get_param_values()

        with auto_load_save_params(
            task=self, auto_read=self._conf_auto_read_params, param_values=param_values
        ):
            result = self.run()

        task_run_executor.sync_local.sync_post_execute()
        # publish all relevant files
        return result

    @property
    def tracker(self):
        return self.current_task_run.tracker

    @property
    def metrics(self):
        # backward compatible code
        return self.tracker

    def get_template_vars(self):
        # TODO: move to cached version, (after relations are built)
        base = {
            "task": self,
            "task_family": self.task_family,
            "task_name": self.task_name,
            "task_signature": self.task_signature,
            "task_id": self.task_id,
        }
        base.update(self._params.get_params_serialized(ParameterFilters.INPUTS))
        if self.task_target_date is None:
            base["task_target_date"] = "input"
        return base

    def on_kill(self):
        """
        Override this method to cleanup subprocesses when a task instance gets killed.

        Any use of the threading, subprocess or multiprocessing
        module within an operator needs to be cleaned up or it will leave
        ghost processes behind.
        """

    def _get_task_output_path_format(self, output_mode):
        """
        Defines the format string used to generate all task outputs.

        For example:
           {root}/{env_label}/{task_target_date}/{task_name}/{task_name}{task_class_version}_{task_signature}/{output_name}{output_ext}
        """
        if self.task_output_path_format:
            # explicit input - first priority
            return self.task_output_path_format
        if self._conf__base_output_path_fmt:
            # from class definition
            return self._conf__base_output_path_fmt

        # default behaviour
        if self.task_env.production and output_mode == OutputMode.prod_immutable:
            return self.run_settings.output.path_prod_immutable_task
        return self.run_settings.output.path_task

    def get_target(self, name, config=None, output_ext=None, output_mode=None):
        name = name or "tmp/dbnd-tmp-%09d" % random.randint(0, 999999999)
        config = config or TargetConfig()
        path_pattern = self._get_task_output_path_format(output_mode)

        path = calculate_path(
            task=self,
            name=name,
            output_ext=output_ext,
            is_dir=config.folder,
            path_pattern=path_pattern,
        )

        return target(path, config=config)

    def get_root(self):
        return self.task_env.root

    def _initialize(self):
        super(Task, self)._initialize()
        self.ctrl._initialize_task()

    def _should_run(self):
        if not self.task_enabled:
            return False

        if self.task_env.production:
            return self.task_enabled_in_prod or self.settings.run.enable_prod

        return True

    def _save_param(self, parameter, original_value, current_value):
        # type: (ParameterDefinition, Any, Any) -> None
        # it's output! we are going to save it.
        # task run doesn't always exist
        task_run = try_get_current_task_run()
        access_status = DbndTargetOperationStatus.OK
        try:
            if isinstance(original_value, InMemoryTarget):
                parameter.value_type = get_value_type_of_obj(
                    current_value, parameter.value_type
                )

            parameter.dump_to_target(original_value, current_value)
            # it's a workaround, we don't want to change parameter for outputs (dynamically)
            # however, we need proper value type to "dump" preview an other meta.
            # we will update it only for In memory targets only for now

        except Exception as ex:
            access_status = DbndTargetOperationStatus.NOK
            raise errors.task_execution.failed_to_save_value_to_target(
                ex, self, parameter, original_value, current_value
            )
        finally:
            if task_run:
                try:
                    task_run.tracker.log_parameter_data(
                        parameter=parameter,
                        target=original_value,
                        value=current_value,
                        operation_type=DbndTargetOperationType.write,
                        operation_status=access_status,
                    )
                except Exception as ex:
                    logger.warning("Failed to log target to tracking store. %s", ex)

    @dbnd_handle_errors(exit_on_error=False)
    def dbnd_run(self):
        # type: (...)-> DatabandRun
        """Run task via Databand execution system."""
        # this code should be executed under context!
        from dbnd._core.current import get_databand_context

        ctx = get_databand_context()
        run = ctx.dbnd_run_task(self)
        return run


@contextmanager
def auto_load_save_params(task, auto_read, param_values):
    original_values = [(p, p.value) for p in param_values]

    # keep all outputs that were "read" during pre-read phase
    # we will not save this values on post-execute phase
    outputs_as_simple_objects = {}

    if auto_read:
        # in .run() we are going to pre-fetch all values
        # * we do it eager loading for all parameters (not lazy)

        for param_value in param_values:
            runtime_value = param_value.parameter.calc_runtime_value(
                param_value.value, task=task
            )
            if param_value.value is not runtime_value:  # only if different
                if param_value.parameter.is_output():
                    # outputs() are going to be processed as well,
                    # we might change them to Path from Target
                    # however, we don't need to consider this value  after that
                    outputs_as_simple_objects[param_value.name] = runtime_value
                param_value.update_param_value(runtime_value)

    try:
        yield original_values

        # we are at the .__exit__ of run, let's save all changed fields
        # no exception, still not good as atomic commit for all files, but less "garbage" left in case of failure
        for param_value, original_value in original_values:
            # TODO: implement Atomic commit
            if not param_value.parameter.is_output():
                continue
            current_value = param_value.value
            if current_value is original_value:
                # nothing to do original_value is the same
                continue
            param_name = param_value.name
            if (
                param_name not in outputs_as_simple_objects
                or outputs_as_simple_objects.get(param_value.name) is not current_value
            ):
                # only if it's user assigned value
                # otherwise we will save Path/PathStr object to the output
                task._save_param(param_value.parameter, original_value, current_value)
            param_value.update_param_value(
                original_value
            )  # do it , so we know what we have processed
    finally:
        # we always want to revert values on .run() exit ( all read and outputs)
        for param_value, original_value in original_values:
            if param_value.value is not original_value:  # only if different
                param_value.update_param_value(original_value)


Task.task_definition.hidden = True

TASK_PARAMS_COUNT = len(
    [
        v
        for k, v in Task.__dict__.items()
        if isinstance(v, ParameterDefinition) and v.value_type_str != "Target"
    ]
)


class TaskCtrl(_BaseTaskCtrl):
    def __init__(self, task):
        super(TaskCtrl, self).__init__(task)

        self.task: Task = task
        from dbnd_run.task_ctrl.task_dag_describe import DescribeDagCtrl
        from dbnd_run.task_ctrl.task_relations import TaskRelations  # noqa: F811
        from dbnd_run.task_ctrl.task_repr import TaskCallRepresentation

        self._should_run = self.task._should_run()
        self._relations = TaskRelations(task)
        self.describe_dag = DescribeDagCtrl(task)
        self.task_repr = TaskCallRepresentation(self.task)

    @property
    def relations(self):  # type: () -> TaskRelations
        return self._relations

    @property
    def task_env(self):
        # type: ()-> EnvConfig
        return self.task.task_env

    def _initialize_task(self):
        # target driven relations are relevant only for orchestration tasks
        self.relations.initialize_relations()

        self.task_dag.initialize_dag_node()

        self.task_repr.initialize()

        super(TaskCtrl, self)._initialize_task()

        # validate circle dependencies
        # may be we should move it to global level because of performance issues
        # however, by running it at every task we'll be able to find the code that causes the issue
        # and show it to user
        if self.task.run_settings.output.recheck_circle_dependencies:
            tasks = self.task_dag.subdag_tasks()
            topological_sort(tasks)

    def should_run(self):
        # convert to property one day
        return self._should_run

    def subdag_tasks(self):
        return self.task_dag.subdag_tasks()

    def save_task_band(self):
        if self.task.task_band:
            task_outputs = traverse_to_str(self.task.task_outputs)
            self.task.task_band.as_object.write_json(task_outputs)

    @contextlib.contextmanager
    def task_context(self, phase):
        # we don't want logs/user wrappers at this stage
        with nested(
            task_context(self.task, phase),
            TaskContextFilter.task_context(self.task.task_id),
            dbnd_config.config_layer_context(self.task.task_config_layer),
        ):
            yield

    @property
    def task_inputs(self):
        return self.relations.task_inputs

    @property
    def task_outputs(self):
        return self.relations.task_outputs

    def banner(self, msg, color=None, task_run=None, verbose=False, exc_info=None):
        from dbnd_run.task_ctrl.task_visualizer import _TaskOrchestrationBannerBuilder

        builder = _TaskOrchestrationBannerBuilder(
            task=self.task,
            msg=msg,
            color=color,
            verbose=verbose,
            print_task_band=self.dbnd_context.run_settings.describe.print_task_band,
        )

        return builder.build_orchestration_banner(task_run=task_run, exc_info=exc_info)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.constants import TaskType
from dbnd_run.task.task import Task


class PythonTask(Task):
    """
    Executes Python code as a dbnd task.

    Example::

        class CalculateAlpha(PythonTask):
            alpha = parameter.value(0.5)
            result = output

            def run(self):
                self.result.write(self.alpha)

    """

    _conf__task_type_name = TaskType.python



# © Copyright Databand.ai, an IBM Company 2022

import logging
import time

from collections import defaultdict

from dbnd._core.errors import DatabandConfigError, friendly_error
from dbnd._core.utils import json_utils
from dbnd._core.utils.traversing import flatten
from dbnd_run.task_ctrl import _TaskRunExecutorCtrl


# Time to sleep while waiting for eventual consistency to finish.
EVENTUAL_CONSISTENCY_SLEEP_INTERVAL = 1

# Maximum number of sleeps for eventual consistency.
EVENTUAL_CONSISTENCY_MAX_SLEEPS = 30


class TaskValidator(_TaskRunExecutorCtrl):
    def find_and_raise_missing_inputs(self):
        missing = find_non_completed(self.relations.task_outputs_user)
        missing_str = non_completed_outputs_to_str(missing)
        raise DatabandConfigError(
            "You are missing some input tasks in your pipeline! \n\t%s\n"
            "The task execution was disabled for '%s'."
            % (missing_str, self.task.task_id)
        )

    def validate_task_inputs(self):
        if not self.task.ctrl.should_run():
            missing = find_non_completed(self.relations.task_outputs_user)
            missing_str = non_completed_outputs_to_str(missing)
            raise DatabandConfigError(
                "You are missing some input tasks in your pipeline! \n\t%s\n"
                "The task execution was disabled for '%s'."
                % (missing_str, self.task.task_id)
            )

        missing = []
        for partial_output in flatten(self.task.ctrl.relations.task_inputs_user):
            if not partial_output.exists():
                missing.append(partial_output)
        if missing:
            raise friendly_error.task_data_source_not_exists(
                self, missing, downstream=[self.task]
            )

    def validate_task_is_complete(self):
        if self.task._complete():
            return

        if self.wait_for_consistency():
            return

        missing = find_non_completed(self.ctrl.relations.task_outputs_user)
        if not missing:
            raise friendly_error.task_has_not_complete_but_all_outputs_exists(self)

        missing_str = non_completed_outputs_to_str(missing)
        raise friendly_error.task_has_missing_outputs_after_execution(
            self.task, missing_str
        )

    def wait_for_consistency(self):
        for attempt in range(EVENTUAL_CONSISTENCY_MAX_SLEEPS):
            missing = find_non_completed(self.task.task_outputs)
            if not missing:
                return True
            missing_and_not_consistent = find_non_consistent(missing)
            if not missing_and_not_consistent:
                return False

            missing_str = non_completed_outputs_to_str(missing_and_not_consistent)
            logging.warning(
                "Some outputs are missing, potentially due to eventual consistency of "
                "your data store. Waining %s second to retry. Additional attempts: %s\n\t%s"
                % (
                    str(EVENTUAL_CONSISTENCY_SLEEP_INTERVAL),
                    str(EVENTUAL_CONSISTENCY_MAX_SLEEPS - attempt),
                    missing_str,
                )
            )

            time.sleep(EVENTUAL_CONSISTENCY_SLEEP_INTERVAL)
        return False


def find_non_completed(targets):
    missing = defaultdict(list)
    for k, v in targets.items():
        for partial_output in flatten(v):
            if not partial_output.exists():
                missing[k].append(partial_output)

    return missing


def find_non_consistent(targets):
    non_consistent = list()
    for k, v in targets.items():
        for partial_output in flatten(v):
            if not partial_output.exist_after_write_consistent():
                non_consistent.append(partial_output)

    return non_consistent


def non_completed_outputs_to_str(non_completed_outputs):
    return json_utils.dumps(non_completed_outputs)



import typing

# © Copyright Databand.ai, an IBM Company 2022
from dbnd._core.task_run.task_run_ctrl import TaskRunCtrl
from dbnd_run.run_settings.env import EnvConfig


if typing.TYPE_CHECKING:

    from dbnd._core.task_run.task_run import TaskRun
    from dbnd_run.task.task import Task


class _TaskRunExecutorCtrl(TaskRunCtrl):
    def __init__(self, task_run: "TaskRun"):
        super().__init__(task_run=task_run)
        self.task: Task = task_run.task

    @property
    def run_executor(self):
        return self.run.run_executor

    @property
    def task_run_executor(self):
        return self.task_run.task_run_executor

    @property
    def task_env(self) -> EnvConfig:
        return self.task.task_env



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd._core.constants import TaskType
from dbnd._core.current import current_task_run
from dbnd._core.task_build.task_context import (
    TaskContextPhase,
    current_phase,
    try_get_current_task,
)
from dbnd._core.task_build.task_results import FuncResultParameter
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.utils.callable_spec import args_to_kwargs
from dbnd_run.errors.task_execution import (
    failed_to_assign_result,
    failed_to_process_non_empty_result,
)
from dbnd_run.task.pipeline_task import PipelineTask
from dbnd_run.task.python_task import PythonTask
from dbnd_run.task.task import Task
from targets.inline_target import InlineTarget


logger = logging.getLogger(__name__)


class _DecoratedCallableTask(Task):
    _dbnd_decorated_task = True

    # default result for any decorated function
    result = None

    def _invoke_func(self, extra_kwargs=None, force_invoke=False):
        # this function is in charge of calling user defined code (decorated function) call
        # usually it's called from from task.run/task.band
        extra_kwargs = extra_kwargs or {}
        spec = self.task_decorator.get_callable_spec()
        invoke_kwargs = {}
        for name in spec.args:
            # if there is no parameter - it was disabled at TaskDefinition building stage
            if self._params.get_param(name) is None:
                continue
            invoke_kwargs[name] = getattr(self, name)

        invoke_kwargs.update(extra_kwargs)

        if not self._dbnd_call_state:
            from dbnd._core.task_build.task_cls__call_state import TaskCallState

            self._dbnd_call_state = TaskCallState()
        self._dbnd_call_state.start()

        if self.task_decorator.is_class:
            # this is the case of
            # @task
            # class UserClass:
            #     pass
            # now we are in the Task instance, it was created via UserClass() at @pipeline
            obj_cls = self.task_decorator.class_or_func
            invoke_kwargs["__call_original_cls"] = True
            self.task_user_obj = obj_cls(**invoke_kwargs)
            result = self.task_user_obj.run()
        else:
            # we are going to run user function
            func_call = spec.item
            result = func_call(**invoke_kwargs)

        self._dbnd_call_state.finish(result)

        result_param = self.__class__.result
        if result_param is None and result:
            raise failed_to_process_non_empty_result(self, result)

        if isinstance(result_param, FuncResultParameter):
            # if we have result that combined from different output params
            # assign all returned values to relevant outputs
            # so they will be automatically saved
            if result is None:
                raise failed_to_assign_result(self, result_param)
            result_param.validate_result(result)
            for r_name, value in result_param.named_results(result):
                setattr(self, r_name, value)
        else:
            self.result = result
        return result

    def on_kill(self):
        task_user_obj = getattr(self, "task_user_obj", None)
        if task_user_obj is not None and hasattr(task_user_obj, "on_kill"):
            task_user_obj.on_kill()
            return
        else:
            super(_DecoratedCallableTask, self).on_kill()


class DecoratedPythonTask(PythonTask, _DecoratedCallableTask):
    _conf__task_type_name = TaskType.python

    def run(self):
        self._invoke_func(force_invoke=True)


class DecoratedPipelineTask(PipelineTask, _DecoratedCallableTask):
    _conf__task_type_name = TaskType.pipeline

    def band(self):
        return self._invoke_func()


def call_decorated_callable_task_with_dbnd_run(
    self: "TaskDecorator", call_args, call_kwargs
):
    # we are at orchestration mode
    task_cls = self.get_task_cls()

    current = try_get_current_task()
    ######
    # current is not None, and we are not in tracking/airflow/luigi
    # this is DBND Orchestration mode
    # we can be in the context of task.run() or in task.band()
    # called from user code using user_decorated_func()  or UserDecoratedTask()

    if self.is_class:
        call_kwargs.pop("__call_original_cls", False)

    # we should not get here from _TaskFromTaskDecorator.invoke()
    # at that function we should call user code directly
    phase = current_phase()
    if phase is TaskContextPhase.BUILD:
        # we are in the @pipeline.band() context, we are building execution plan
        t = task_cls(*call_args, **call_kwargs)

        # we are in the band, and if user_code() is called we want to remove redundant
        # `user_code().result` usage
        if t.task_definition.single_result_output:
            return t.result

        # we have multiple outputs (more than one "output" parameter)
        # just return task object, user will use it as `user_code().output_1`
        return t
    elif phase is TaskContextPhase.RUN:
        # we are "running" inside some other task execution (orchestration!)
        #  (inside user_defined_function() or UserDefinedTask.run()

        from dbnd_run.current import try_get_run_executor

        run_executor = try_get_run_executor()
        if not run_executor:
            raise Exception("Run Executor is not activated")
        # if possible we will run it as "orchestration" task
        # with parameters parsing
        if (
            run_executor.run_config.task_run_at_execution_time_enabled
            and current.task_supports_dynamic_tasks
        ):
            return _run_task_from_another_task_execution(
                self,
                run_executor=run_executor,
                parent_task=current,
                call_args=call_args,
                call_kwargs=call_kwargs,
            )
        # we can not call it in "dbnd" way, fallback to normal call
        if self.is_class:
            call_kwargs["__call_original_cls"] = False
        return self.class_or_func(*call_args, **call_kwargs)
    else:
        raise Exception()


def _run_task_from_another_task_execution(
    self: "TaskDecorator", run_executor, parent_task: Task, call_args, call_kwargs
) -> TaskRun:
    # task is running from another task
    task_cls = self.get_task_cls()

    from dbnd import PipelineTask, pipeline
    from dbnd._core.task_build.dbnd_decorator import _default_output

    # orig_call_args, orig_call_kwargs = call_args, call_kwargs
    call_args, call_kwargs = args_to_kwargs(
        self.get_callable_spec().args, call_args, call_kwargs
    )

    # Map all kwargs to the "original" target of that objects
    # for example: for DataFrame we'll try to find a relevant target that were used to read it
    # get all possible value's targets
    call_kwargs_as_targets = run_executor.run.target_origin.get_for_map(call_kwargs)
    for p_name, value_origin in call_kwargs_as_targets.items():
        root_target = value_origin.origin_target
        path = root_target.path if hasattr(root_target, "path") else None
        original_object = call_kwargs[p_name]
        call_kwargs[p_name] = InlineTarget(
            root_target=root_target,
            obj=original_object,
            value_type=value_origin.value_type,
            source=value_origin.origin_target.source,
            path=path,
        )

    call_kwargs.setdefault("task_is_dynamic", True)
    call_kwargs.setdefault(
        "task_in_memory_outputs",
        run_executor.run_config.task_run_at_execution_time_in_memory_outputs,
    )

    if issubclass(task_cls, PipelineTask):
        # if it's pipeline - create new databand run
        # create override _task_default_result to be object instead of target
        task_cls = pipeline(
            self.class_or_func, _task_default_result=_default_output
        ).task_cls

        # instantiate new inline execution
        task = task_cls(*call_args, **call_kwargs)
        # if it's pipeline - create new databand run
        run = run_executor.databand_context.dbnd_run_task(task)
        task_run = run.get_task_run(task.task_id)
    else:
        # instantiate inline task (dbnd object)
        task = task_cls(*call_args, **call_kwargs)

        # update upstream/downstream relations - needed for correct tracking
        # we can have the task as upstream , as it was executed already
        if not parent_task.task_dag.has_upstream(task):
            parent_task.set_upstream(task)

        from dbnd._core.task_build.task_cls__call_state import TaskCallState

        task._dbnd_call_state = TaskCallState(should_store_result=True)
        try:
            parent_task_run = current_task_run()
            task_run = run_executor.run_task_at_execution_time(
                task, task_engine=parent_task_run.task_run_executor.task_engine
            )

            # this will work only for _DecoratedTask
            if task._dbnd_call_state.result_saved:
                return task._dbnd_call_state.result

        finally:
            # we'd better clean _invoke_result to avoid memory leaks
            task._dbnd_call_state = None

    # if we are inside run, we want to have real values, not deferred!
    if task.task_definition.single_result_output:
        return task.__class__.result.load_from_target(task.result)
        # we have func without result, just fallback to None
    return task



# © Copyright Databand.ai, an IBM Company 2022

import logging
import os
import shlex
import signal
import subprocess

from subprocess import list2cmdline
from typing import Any, Dict, List, Optional

import six

from dbnd import log_metric, parameter, task
from dbnd._core.current import try_get_current_task_run
from dbnd._core.errors import DatabandConfigError
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.utils.basics.signal_utils import safe_signal
from dbnd._core.utils.platform import windows_compatible_mode
from dbnd_run.errors.task_execution import failed_to_run_cmd


logger = logging.getLogger(__name__)


@task(popen_kwargs=parameter(empty_default=True)[Dict[str, str]])
def bash_cmd(
    cmd=None,
    args=None,
    check_retcode=0,
    cwd=None,
    env=None,
    dbnd_env=True,
    output_encoding="utf-8",
    popen_kwargs=None,
    wait_for_termination_s=5,
    shell=False,
):
    # type:( str, List[str], Optional[int], str, Dict[str,str], bool, str, Dict[str,Any]) -> int
    if popen_kwargs is None:
        popen_kwargs = dict()
    popen_kwargs = popen_kwargs.copy()

    if cmd and args:
        raise DatabandConfigError("You should not provide cmd and args ")

    if cmd:
        if shell:
            args = cmd
        else:
            args = shlex.split(cmd)
    elif args:
        args = list(map(str, args))
        cmd = list2cmdline(args)
        if shell:
            args = cmd

    logger.info("Running: " + cmd)  # To simplify rerunning failing tests

    if dbnd_env and DatabandRun.has_instance():
        env = env or os.environ.copy()
        dbnd_env_vars = DatabandRun.get_instance().get_context_spawn_env()
        logger.info(
            "Exporting the following env vars:\n%s",
            "\n".join(["{}={}".format(k, v) for k, v in dbnd_env_vars.items()]),
        )
        env.update()

    def preexec_fn():
        if windows_compatible_mode:
            return
        # Restore default signal disposition and invoke setsid
        for sig in ("SIGPIPE", "SIGXFZ", "SIGXFSZ"):
            if hasattr(signal, sig):
                safe_signal(getattr(signal, sig), signal.SIG_DFL)
        os.setsid()

    process = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=-1,
        universal_newlines=True,
        env=env,
        preexec_fn=preexec_fn,
        cwd=cwd,
        shell=shell,  # nosec B602 (Ignore this for bandit as this is code used for orchestration)
        **popen_kwargs
    )

    try:
        task_run = try_get_current_task_run()
        if task_run:
            task_run.task.process = process

        logger.info("Process is running, output:")
        # While command is running let read it's output
        output = []
        while True:
            line = process.stdout.readline()
            if line == "" or line == b"":
                break
            line = safe_decode(line, output_encoding).rstrip()

            logger.info("out: %s", line)
            # keep last 1000 lines only
            output.append(line)
            if len(output) > 1500:
                output = output[-1000:]

        returncode = process.wait()
        logger.info("Command exited with return code %s", process.returncode)
        if check_retcode is not None and returncode != check_retcode:
            raise failed_to_run_cmd(
                "Bash command failed", cmd_str=cmd, return_code=returncode
            )
        return returncode
    except Exception:
        logger.info("Received interrupt. Terminating subprocess and waiting")
        try:
            process.terminate()
            process.wait(wait_for_termination_s)
        except Exception:
            pass
        raise


def safe_decode(line, output_encoding):
    if isinstance(line, six.string_types):
        return line
    return line.decode(output_encoding)


@task(popen_kwargs=parameter(empty_default=True)[Dict[str, str]])
def bash_script(
    script=None,
    check_retcode=0,
    cwd=None,
    env=None,
    dbnd_env=True,
    output_encoding="utf-8",
    popen_kwargs=None,
):
    # type:( str, Optional[int],str, Dict[str,str], bool, str, Dict[str,Any]) -> int

    # we need a working folder to create bash script
    task_run = try_get_current_task_run()
    if task_run:
        script_dir = str(task_run.task_run_executor.local_task_run_root)
    else:
        script_dir = None

    bash_script_path = os.path.join(script_dir, "bash_cmd.sh")
    with open(bash_script_path, "wb") as bs:
        bs.write(bytes(script, "utf_8"))

    log_metric("bash_script", bash_script_path)

    logger.info("Bash script location: %s", bash_script_path)
    args = ["bash", bash_script_path]
    return bash_cmd.func(
        args=args,
        check_retcode=check_retcode,
        cwd=cwd or script_dir,
        env=env,
        dbnd_env=dbnd_env,
        output_encoding=output_encoding,
        popen_kwargs=popen_kwargs,
    )



# © Copyright Databand.ai, an IBM Company 2022

import importlib
import logging

from dbnd._core.errors import friendly_error
from dbnd._core.utils.basics.load_python_module import _load_module
from dbnd._core.utils.seven import import_errors
from dbnd._vendor import pluggy
from dbnd_run.plugin import dbnd_plugin_spec


logger = logging.getLogger(__name__)

hookimpl = pluggy.HookimplMarker("dbnd")
pm = pluggy.PluginManager("dbnd")
pm.add_hookspecs(dbnd_plugin_spec)


# all other modules
def is_plugin_enabled(module, module_import=None):
    if pm.has_plugin(module):
        return True

    if module_import:
        try:
            importlib.import_module(module_import)
            return True
        except import_errors:
            pass
    if "-" in module:
        # Workaround for manually loaded plugins (contain `_` instead of `-`)
        # Occurs when 'disable_pluggy_entrypoint_loading' is turned on in spark-config. All plugins are loaded
        # manually (spark_ctrl.py:107), so we must fix `plugin` syntax ('-') to python module syntax ('_')
        return is_plugin_enabled(module.replace("-", "_"), module_import)
    return False


def assert_plugin_enabled(module, reason=None, module_import=None):
    if not is_plugin_enabled(module, module_import=module_import):
        raise friendly_error.config.missing_module(module, reason)
    return True


_dbnd_plugins_registered = False


def register_dbnd_plugins():
    pm.load_setuptools_entrypoints("dbnd")
    pm.check_pending()


def register_dbnd_user_plugins(user_plugin_modules):
    for plugin_module in user_plugin_modules:
        module = _load_module(plugin_module, "plugin:%s" % plugin_module)
        pm.register(module)

        base_msg = "Plugin %s" % plugin_module
        if getattr(module, "__version__", None):
            base_msg += " v%s" % module.__version__

        logger.info(base_msg + " loaded...")



# © Copyright Databand.ai, an IBM Company 2022

from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.errors.friendly_error.helpers import (
    _parameter_name,
    _run_name,
    _task_name,
)


def failed_to_save_value_to_target(ex, task, parameter, target, value):
    if parameter.name == "result":
        return DatabandRuntimeError(
            "Can't save return value of %s to %s': %s"
            % (_parameter_name(task, parameter), target, ex),
            show_exc_info=True,
            nested_exceptions=[ex],
            help_msg="Check your {} return value "
            "and it definition ('{}')".format(
                _run_name(task), parameter.value_type.type
            ),
        )
    return DatabandRuntimeError(
        "Can't save %s to %s': %s" % (_parameter_name(task, parameter), target, ex),
        show_exc_info=True,
        nested_exceptions=[ex],
        help_msg="Check your %s logic. " % task.friendly_task_name,
    )


def failed_to_read_value_from_target(ex, task, parameter, target):
    return DatabandRuntimeError(
        "Can't read %s from %s': %s" % (_parameter_name(task, parameter), target, ex),
        show_exc_info=True,
        nested_exceptions=[ex],
        help_msg="Check your %s logic. " % task.friendly_task_name,
    )


def failed_to_assign_result(task, result_parameter):
    return DatabandRuntimeError(
        "The result of the band/run call is None, "
        "it can not be assigned to {schema}".format(
            task=task, schema=result_parameter.schema
        ),
        help_msg="Check your %s return value" % (_task_name(task)),
    )


def failed_to_process_non_empty_result(task, result):
    return DatabandRuntimeError(
        "Can' process non empty result of {task} while it's marked as task without outputs: result={result}".format(
            task=_task_name(task), result=result
        ),
        help_msg="Please, use @task(result=YOU RESULT SCHEMA)",
    )


def can_run_only_tasks(task):
    return DatabandRuntimeError(
        "Databand can run only Tasks, got {task} instead".format(task=type(task)),
        help_msg="Please, use check that you don't call function while providing it to databand. "
        "Use YOUR_TASK_FUNCTION.task()",
    )


def wrong_return_value_type(task_def, names, result):
    return DatabandRuntimeError(
        "Returned value from '{task}' should be tuple/list/dict as task has multiple result."
        "Expected tuple of '{names}', got value of type '{result}'".format(
            task=task_def.run_name(), names=names, result=type(result)
        )
    )


def wrong_return_value_len(task_def, names, result):
    return DatabandRuntimeError(
        "Returned result from '{task}' doesn't match expected schema. "
        "Expected tuple of '{names}', got tuple of length '{result}'".format(
            task=task_def.run_name(), names=names, result=len(result)
        )
    )


def failed_to_run_spark_script(task, cmd, application, return_code, error_snippets):
    return DatabandRuntimeError(
        "spark_submit failed with return code %s. Failed to run: %s"
        % (return_code, cmd),
        show_exc_info=False,
        nested_exceptions=error_snippets,
        help_msg="Check your  %s logic and input data %s. Inspect spark logs for more info."
        % (application, list(task.relations.task_inputs_user.values())),
    )


def failed_to_run_cmd(name, cmd_str, return_code):
    return DatabandRuntimeError(
        "{name} has failed, returncode='{return_code}'. Failed to run: {cmd}".format(
            name=name, return_code=return_code, cmd=cmd_str
        ),
        show_exc_info=False,
        help_msg="Inspect logs for more info.",
    )


def failed_spark_status(msg):
    return DatabandRuntimeError(msg, show_exc_info=False)


def failed_to_run_emr_step(reason, logs_path, error_snippets):
    if logs_path and error_snippets:
        return DatabandRuntimeError(
            "EMR Spark step failed with reason: %s " % reason,
            show_exc_info=False,
            nested_exceptions=error_snippets,
            help_msg="Check your application logic. Inspect spark emr logs for more info.\n "
            "Logs are available at %s." % logs_path,
        )
    return DatabandRuntimeError(
        "EMR Spark step failed with reason: %s. Additionally Databand failed to get EMR logs."
        % reason,
        show_exc_info=False,
        help_msg="Check your application logic. Inspect emr console for logs and more info\n ",
    )


def system_exit_at_task_run(task, ex):
    return DatabandRuntimeError(
        "Task execution has been aborted with sys.exit() call: %s" % ex,
        nested_exceptions=ex,
        show_exc_info=False,
        help_msg="Check your task run()\n ",
    )


def databand_context_killed(interrupt_location):
    return DatabandRuntimeError(
        "Databand Context has been killed externaly. Interrupted at %s",
        interrupt_location,
    )



# © Copyright Databand.ai, an IBM Company 2022
import typing

from dbnd._core.run.run_ctrl import RunCtrl


if typing.TYPE_CHECKING:
    from typing import List

    from dbnd._core.task_run.task_run import TaskRun
    from dbnd_run.run_executor.run_executor import RunExecutor
    from dbnd_run.run_settings import EngineConfig


class RunExecutorEngine(RunCtrl):
    def __init__(
        self,
        run_executor: "RunExecutor",
        task_executor_type: str,
        host_engine: "EngineConfig",
        target_engine: "EngineConfig",
        task_runs: "List[TaskRun]",
    ):
        super(RunExecutorEngine, self).__init__(run=run_executor.run)
        self.run_executor = run_executor
        self.task_executor_type = task_executor_type
        self.host_engine = host_engine
        self.target_engine = target_engine
        self.task_runs = task_runs

    def do_run(self):
        raise NotImplementedError()



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import print_function

import logging

from dbnd._core.constants import DescribeFormat
from dbnd._core.errors import DatabandSystemError
from dbnd._core.task_ctrl.task_ctrl import TaskSubCtrl
from dbnd._core.utils.basics.helpers import indent
from dbnd._vendor.termcolor import colored
from dbnd_run.run_settings import DescribeConfig
from dbnd_run.tasks import DataSourceTask


logger = logging.getLogger(__name__)


def tasks_trail(tasks):
    task_ids = [t.task_id for t in tasks]
    return " -> ".join(task_ids)


class DescribeDagCtrl(TaskSubCtrl):
    def __init__(self, task, describe_format=DescribeFormat.long, complete_status=None):
        super(DescribeDagCtrl, self).__init__(task)

        self.describe_format = describe_format

        # dummy implementation of complete cache
        self._complete_status = complete_status or {}

    @property
    def config(self):
        return self.dbnd_context.run_settings.describe

    def tree_view(self, describe_format=None):
        """
        Shows an ascii tree representation of the DAG
        """
        # TODO: change to treelib implementation

        seen = set()

        def get_downstream(task, level=0):
            task_desc = self._describe_task(task, describe_format=describe_format)
            if task in seen:
                return [(level, "%s (*)" % task_desc)]
            result = [(level, task_desc)]

            seen.add(task)
            level += 1
            count = 0
            task_dag = task.ctrl.task_dag
            for t in task_dag.upstream:
                count += 1
                if isinstance(t, DataSourceTask):
                    continue

                if count > 30:
                    result.append((level, "..(%s tasks).." % len(task_dag.upstream)))
                    break
                result.extend(get_downstream(t, level))

            return result

        result = get_downstream(self.task)

        messages = [indent(msg, "\t" * level) for level, msg in result]
        logger.info(
            "Tasks Graph - (*) represent existing node in the graph run "
            "(green is completed, yellow is going to be submitted):\n%s",
            "\n".join(messages),
        )

    def list_view(self):
        logger.info("List View of the DAG:\n")
        for t in self.task_dag.subdag_tasks():
            logger.info("%s\n" % self._describe_task(t))

    def _get_task_complete(self, task):
        if task.task_id not in self._complete_status:
            try:
                complete = task._complete()
            except Exception as ex:
                logger.warning(
                    "Failed to get complete status for %s: %s", task.task_id, ex
                )
                complete = None

            self._complete_status[task.task_id] = complete
        return self._complete_status[task.task_id]

    def _describe_task(self, task, describe_format=None, msg=None, color=None):
        describe_format = describe_format or self.describe_format
        describe_config = self.config  # type: DescribeConfig
        msg = msg or ""

        if color is None:
            color = "white"
            if not describe_config.no_checks:
                color = "green" if self._get_task_complete(task) else "cyan"

        if describe_format == DescribeFormat.short:
            return colored(str(task.task_id), color)

        if (
            describe_format == DescribeFormat.long
            or describe_format == DescribeFormat.verbose
        ):
            title = "%s - %s" % (task.task_name, task.task_id)
            if task.task_name != task.get_task_family():
                title += "(%s)" % task.get_task_family()
            if msg:
                title += ": %s" % msg
            return task.ctrl.banner(
                title, color=color, verbose=describe_format == DescribeFormat.verbose
            )

        raise DatabandSystemError("Not supported format mode %s" % self.describe_format)

    def describe_dag(self):
        # print short tree
        self.tree_view(describe_format=DescribeFormat.short)
        self.list_view()

    def describe(self, as_tree=False):
        if as_tree:
            from dbnd._core.constants import DescribeFormat

            self.tree_view(describe_format=DescribeFormat.short)
        else:
            self.ctrl.describe_dag.list_view()


def print_tasks_tree(root_task, task_runs, describe_format=DescribeFormat.short):
    from dbnd_run.task_ctrl.task_dag_describe import DescribeDagCtrl

    completed = {tr.task.task_id: tr.is_reused for tr in task_runs}
    run_describe_dag = DescribeDagCtrl(
        root_task, describe_format, complete_status=completed
    )
    run_describe_dag.tree_view(describe_format=describe_format)



# © Copyright Databand.ai, an IBM Company 2022

import hashlib
import logging

from os import path

import six

from dbnd._core.utils.task_utils import targets_to_str
from dbnd_run.task_ctrl import _TaskRunExecutorCtrl
from targets import Target, target
from targets.fs import FileSystems


logger = logging.getLogger(__name__)


class TaskSyncCtrl(_TaskRunExecutorCtrl):
    def __init__(self, task_run):
        super(TaskSyncCtrl, self).__init__(task_run=task_run)

        self.remote_sync_root = self.task.task_env.dbnd_data_sync_root.folder("deploy")

    def sync_files(self, local_files):
        if not local_files:
            return []
        return [self.sync(f) for f in local_files]

    def sync(self, local_file):
        if not local_file:
            #  should return None, not empty string to be compatible with airflow code
            return None

        if not isinstance(local_file, Target):
            local_file = target(local_file)
        return str(self._sync(local_file))

    def arg_files(self, local_files):
        #  should return None, not empty string to be compatible with airflow code
        if not local_files:
            return None
        if isinstance(local_files, six.string_types):
            local_files = [local_files]
        synced_files = self.sync_files(local_files)
        synced_files_str = targets_to_str(synced_files)
        return ",".join(synced_files_str)

    def is_remote(self, file):
        if not file:
            return None
        return file.fs.name != FileSystems.local

    def _md5(self, local_path):
        with open(local_path, "rb+") as f:
            checksum = hashlib.md5(f.read())  # nosec B324
        return checksum.hexdigest()

    def remote_file(self, local_file, md5_hash):
        if self.is_remote(local_file):
            return local_file

        file_name = path.basename(local_file.path)
        remote_file_name = "{}/{}".format(md5_hash, file_name)
        return self.remote_sync_root.partition(remote_file_name)

    def _sync_remote(self, remote_file):
        """
        Synchronizes remote fs -> local fs
        """
        return remote_file

    def _sync_local(self, local_file):
        """
        Synchronizes local fs -> remote fs
        """
        md5_hash = self._md5(local_file.path)
        remote_file = self.remote_file(local_file, md5_hash)

        if not remote_file.exists():
            try:
                self._upload(local_file, remote_file)
            except:
                if not remote_file.exists():
                    logger.error(
                        "Unexpected error occurred while trying to upload file `%s` to the remote server",
                        remote_file,
                    )
                    raise
                logger.warning(
                    "File %s was copied from different process to the remote location, skipping upload...",
                    remote_file,
                )

        else:
            logger.info(
                "File %s already exists at remote location, skipping upload...",
                remote_file,
            )

        return remote_file

    def _sync(self, local_file):
        if self.is_remote(local_file):
            remote_file = self._sync_remote(local_file)
        else:
            remote_file = self._sync_local(local_file)
        return remote_file

    def _upload(self, local_file, remote_file):
        remote_file.copy_from_local(local_file.path)

    def _exists(self, remote_file):
        return remote_file.exists()


class DisabledTaskSyncCtrl(TaskSyncCtrl):
    def _sync(self, local_file):
        return local_file
